from typing import Any, Sequence, Union
from cachelm.vectorizers.vectorizer import Vectorizer

try:
    from fastembed import TextEmbedding
except ImportError:
    raise ImportError(
        "FastEmbed library is not installed. Run `pip install fastembed` to install it."
    )


class FastEmbedVectorizer(Vectorizer):
    """
    FastEmbed embedding model.
    """

    def __init__(
        self,
        model_name: str = "BAAI/bge-base-en",
        cache_dir: str | None = None,
        threads: int | None = None,
        providers: Sequence[Union[str, tuple[str, dict[Any, Any]]]] | None = None,
        cuda: bool = False,
        device_ids: list[int] | None = None,
        lazy_load: bool = False,
    ):
        """
        Initialize the FastEmbed embedding model.
        Args:
            model_name (str): The name of the FastEmbed model to use.
            cache_dir (str | None): The directory to cache the model.
            threads (int | None): The number of threads to use for embedding.
            providers (Sequence[Union[str, tuple[str, dict[Any, Any]]]] | None): The providers to use for embedding.
            cuda (bool): Whether to use CUDA for embedding.
            device_ids (list[int] | None): The device IDs to use for embedding.
            lazy_load (bool): Whether to lazy load the model.
        """
        super().__init__()
        self.embedding_model = TextEmbedding(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=threads,
            providers=providers,
            cuda=cuda,
            device_ids=device_ids,
            lazy_load=lazy_load,
        )

    def embed(self, text):
        """
        Embed the chat history.
        """
        out = list(self.embedding_model.embed(text))[0].tolist()
        return out

    def embed_many(self, text: list[str]) -> list[list[float]]:
        """
        Embed the chat history.
        """
        outs = self.embedding_model.embed(text)
        return [o.tolist() for o in outs]



from abc import ABC, abstractmethod
from typing import TypeVar, Generic
from cachelm.databases.database import Database
from loguru import logger
import signal

from cachelm.middlewares.deduper import Deduper
from cachelm.middlewares.middleware import Middleware
from cachelm.types.chat_history import ChatHistory, Message

T = TypeVar("T")


class Adaptor(ABC, Generic[T]):
    """
    Base class for all adaptors.
    """

    def __init__(
        self,
        module: T,
        database: Database,
        window_size: int = 2,
        distance_threshold: float = 0.2,
        dispose_on_sigint: bool = False,
        middlewares: list[Middleware] = [],
        dedupe: bool = True,
    ):
        """
        Initialize the adaptor with a module, database, and configuration options.

        Args:
            module: The module to be adapted.
            database: The database instance used for caching.
            window_size: Number of recent messages to consider for caching (default: 4).
            distance_threshold: Similarity threshold for cache retrieval (default: 0.2).
            dispose_on_sigint: If True, dispose adaptor on SIGINT signal (default: False).
            middlewares: List of middlewares to apply to the messages (default: empty list).
            dedupe: If True, apply deduplication middleware (default: True).
        """
        self._validate_inputs(database, window_size, distance_threshold)
        self._initialize_attributes(
            module, database, window_size, distance_threshold, middlewares, dedupe
        )
        if dispose_on_sigint:
            signal.signal(signal.SIGINT, self._handle_sigint)

    def _handle_sigint(self, signum, frame):
        """
        Handle SIGINT signal.
        """
        logger.info("SIGINT received, disposing of the adaptor")
        self.dispose()
        exit(0)

    def _validate_inputs(
        self, database: Database, window_size: int, distance_threshold: float
    ):
        """
        Validate the inputs for the adaptor.
        """
        if not isinstance(database, Database):
            raise TypeError("Database must be an instance of Database")
        if distance_threshold < 0 or distance_threshold > 1:
            raise ValueError("Distance threshold must be between 0 and 1")
        if window_size < 0:
            raise ValueError("Window size must be greater than or equal to 0")

    def _initialize_attributes(
        self,
        module: T,
        database: Database,
        window_size: int,
        distance_threshold: float,
        middlewares: list[Middleware],
        dedupe: bool,
    ):
        """
        Initialize the attributes for the adaptor.
        """
        success = database.connect()
        if not success:
            raise Exception("Failed to connect to the database")
        logger.info("Connected to the database")
        self.database = database
        self.module = module
        self.history = ChatHistory()
        self.window_size = window_size
        self.distance_threshold = distance_threshold
        self.middlewares = middlewares
        if dedupe:
            self.middlewares.append(Deduper())

    @abstractmethod
    def get_adapted(self) -> T:
        """
        Get the adapted module.
        """
        raise NotImplementedError("getAdapted method not implemented")

    def set_history(self, messages: list[Message]):
        """
        Set the chat history.
        """
        self.history.set_messages(messages)

    def add_user_message(self, message: Message):
        """
        Add a user message to the chat history.
        """
        self.history.add_user_message(message)

    def add_assistant_message(self, message: Message, save_to_db: bool = True):
        """
        Add an assistant message to the chat history.
        Applies all middlewares to the message (pre-cache)
        """
        lastMessagesWindow = self.history.get_messages(self.window_size)
        for middleware in self.middlewares:
            message = middleware.pre_cache_save(message, self.history)
            if message is None:
                return
        self.database.write(lastMessagesWindow, message)

    def get_cache(self):
        """
        Get the cache from the database.
        Applies all middlewares to the cache (post-cache).

        If the cache is empty, return None.
        If the cache is not empty, add it to the history.

        """
        cache = self.database.find(
            self.history.get_messages(self.window_size),
            self.distance_threshold,
        )
        if not cache:
            return None

        for middleware in self.middlewares:
            cache = middleware.post_cache_retrieval(cache, self.history)
            if cache is None:
                return None
        # Add the cache to the history
        self.history.add_assistant_message(cache)
        return cache

    def dispose(self):
        """
        Dispose of the adaptor.
        """
        self.database.disconnect()
        logger.info("Disconnected from the database")



import unittest
from cachelm.databases.database import Database
from cachelm.types.chat_history import Message


class TestDatabases(unittest.TestCase):
    def _test_helper(self, db: Database):
        """
        Test the database connection and basic operations.
        """
        db.reset()

        # Test writing to the database
        history = [
            Message(role="user", content="Hello, how are you?"),
            Message(role="assistant", content="I'm fine, thank you!"),
            Message(role="user", content="What can you do?"),
        ]
        response = Message(
            role="assistant",
            content="I can assist you with various tasks.",
        )
        db.write(history, response)

        # Test finding from the database
        result = db.find(history)
        assert result is not None, "Find result should not be None"
        assert (
            result.content == response.content
        ), "Find result should match written response"

    def test_chroma_database(self):
        """
        Test the Chroma database.
        """
        from cachelm.databases.chroma import ChromaDatabase
        from cachelm.vectorizers.fastembed import FastEmbedVectorizer

        vectorizer = FastEmbedVectorizer()
        db = ChromaDatabase(vectorizer)
        success = db.connect()
        assert success, "Failed to connect to Chroma database"
        self._test_helper(db)
        db.disconnect()

    def test_clickhouse_database(self):
        """
        Test the ClickHouse database.
        """
        from cachelm.databases.clickhouse import ClickHouse
        from cachelm.vectorizers.fastembed import FastEmbedVectorizer

        vectorizer = FastEmbedVectorizer()
        db = ClickHouse(
            host="localhost",
            port=18123,
            user="default",
            password="pass",
            vectorizer=vectorizer,
        )
        success = db.connect()
        assert success, "Failed to connect to ClickHouse database"
        self._test_helper(db)
        db.disconnect()

    def test_redisvl_database(self):
        """
        Test the RedisVL database.
        """
        from cachelm.databases.redisvl import RedisVLDatabase
        from cachelm.vectorizers.fastembed import FastEmbedVectorizer

        vectorizer = FastEmbedVectorizer()
        db = RedisVLDatabase(host="localhost", port=16379, vectorizer=vectorizer)
        success = db.connect()
        assert success, "Failed to connect to RedisVL database"
        self._test_helper(db)
        db.disconnect()



from cachelm.middlewares.middleware import Middleware
import re

from cachelm.types.chat_history import Message


class Skipper(Middleware):
    """
    Middleware that skips saving messages to cache if they match any of the provided regex patterns.

    This is useful for filtering out messages that should not be cached based on regular expressions.

    Example:
        from cachelm.middlewares.skipper import Skipper

        # Create a Skipper middleware instance with regex patterns to skip
        skipper = Skipper(patterns=[r"skip_this.*", r"ignore_\d+"])
    """

    def __init__(self, patterns: list[str], function_calls: list[str] = []):
        """
        Initialize the Skipper middleware.

        Args:
            patterns: list[str]: A list of Replacement objects.
            function_calls: list[str]: A list of function calls to be skipped/
        """
        self.patterns = patterns

    def pre_cache_save(self, message, history):
        """
        Pre-cache save method to check if the message should be skipped.
        Args:
            message: The message to be checked.
            history: The history of messages.
        Returns:
            None if the message should be skipped, otherwise the message itself.
        """
        for pattern in self.patterns:
            if re.search(pattern, message.content):
                # If the pattern is found in the message content, skip saving it to cache
                return None
        return message

    def _should_skip_function_call(self, message: Message):
        """
        Check if the message is a function call that should be skipped.
        Args:
            message: The message to be checked.
        Returns:
            True if the message is a function call that should be skipped, otherwise False.
        """
        if (
            hasattr(message, "tool_calls")
            and message.tool_calls
            and isinstance(message.tool_calls, list)
        ):
            function_call = message.tool_calls[0]
            if function_call.tool in self.patterns:
                # If the function call matches any of the patterns, skip saving it to cache
                return True
        return False

    def post_cache_retrieval(self, message, history):
        return message



from cachelm.vectorizers.vectorizer import Vectorizer

try:
    from redisvl.utils.vectorize import BaseVectorizer, HFTextVectorizer
except ImportError:
    raise ImportError(
        "RedisVL library is not installed. Run `pip install redisvl` to install it."
    )


class RedisvlVectorizer(Vectorizer):
    """
    RedisVL embedding model.
    """

    def __init__(
        self,
        vectorizer: BaseVectorizer = HFTextVectorizer(
            model="sentence-transformers/all-mpnet-base-v2",
        ),
    ):
        """
        Initialize the RedisVL embedding model.
        Args:
            vectorizer (BaseVectorizer): The RedisVL vectorizer to use.
        """
        super().__init__()
        self.vectorizer = vectorizer

    def embed(self, text):
        """
        Embed the chat history.
        """
        out = self.vectorizer.embed(text)
        return out

    def embed_many(self, text: list[str]) -> list[list[float]]:
        """
        Embed the chat history.
        """
        outs = self.vectorizer.embed_many(text)
        return outs



from typing import Literal
from cachelm.middlewares.middleware import Middleware


class Replacement:
    """
    A class representing a replacement operation.
    """

    def __init__(self, key: str, value: str):
        """
        Initialize the Replacement object.

        Args:
            key (str): The inner representation of the string to be replaced.
            value (str): The string to replace with.
        """
        self.key = key
        self.value = value


class Replacer(Middleware):

    def __init__(self, replacements: list[Replacement]):
        """
        Initialize the Replacer middleware.

        Args:
            replacements: list[Replacement]: A list of Replacement objects.
        """
        self.replacements = replacements

    def pre_cache_save(self, message, history):
        for replacement in self.replacements:
            message.content = message.content.replace(
                replacement.value, replacement.key
            )
        return message

    def post_cache_retrieval(self, message, history):
        for replacement in self.replacements:
            message.content = message.content.replace(
                replacement.key, replacement.value
            )
        return message



from cachelm.vectorizers.vectorizer import Vectorizer

try:
    from chromadb.utils import embedding_functions
except ImportError:
    raise ImportError(
        "ChromaDB library is not installed. Run `pip install chromadb` to install it."
    )


class ChromaVectorizer(Vectorizer):
    """
    ChromaDB embedding function.
    """

    def __init__(
        self,
        vectorizer: embedding_functions.EmbeddingFunction[
            embedding_functions.Documents
        ] = embedding_functions.Text2VecEmbeddingFunction(),
    ):
        """
        Initialize the ChromaDB embedding function
        Args:
            vectorizer (embedding_functions.EmbeddingFunction[Documents]): The ChromaDB vectorizer to use.
        """
        super().__init__()
        if not isinstance(
            vectorizer,
            embedding_functions.EmbeddingFunction,
        ):
            raise TypeError(
                "vectorizer must be an instance of chromadb.EmbeddingFunction[Documents] "
            )
        self.vectorizer = vectorizer

    def embed(self, text):
        """
        Embed the chat history.
        """
        out = self.vectorizer([text])[0]
        return out.tolist()

    def embed_many(self, text: list[str]) -> list[list[float]]:
        """
        Embed the chat history.
        """
        outs = self.vectorizer(text)
        return [o.tolist() for o in outs]



from abc import ABC, abstractmethod


class Vectorizer(ABC):
    """
    Base class for all embedders.
    """

    @abstractmethod
    def embed(self, text: str) -> list[float]:
        """
        Embed a single text string into a vector.
        Args:
            text (str): The text to embed.
        Returns:
            list[float]: The embedded vector.
        """
        raise NotImplementedError("embed method not implemented")

    @abstractmethod
    def embed_many(self, text: list[str]) -> list[list[float]]:
        """
        Embed multiple text strings into vectors.
        Args:
            text (list[str]): The list of texts to embed.
        Returns:
            list[list[float]]: The list of embedded vectors.
        """
        raise NotImplementedError("embed method not implemented")



from loguru import logger
from cachelm.middlewares.middleware import Middleware


class Deduper(Middleware):
    """
    Middleware that returns None if the reply is already present in history.
    """

    def pre_cache_save(self, message, history):
        return message

    def post_cache_retrieval(self, message, history):
        logger.debug(
            f"Deduper: Checking if message is already in history: {message.content}"
        )
        logger.debug(f"Deduper: History: {[m.content for m in history]}")
        # No modification after LLM response
        # Check if message content is already in any previous message in history
        for past_message in history:
            if getattr(past_message, "content", None) == message.content:
                return None
        return message



import os
import time
import asyncio
from cachelm.adaptors.openai import OpenAIAdaptor
from cachelm.databases.clickhouse import ClickHouse
from cachelm.vectorizers.fastembed import FastEmbedVectorizer
from openai import AsyncOpenAI
from cachelm.middlewares.replacer import Replacer, Replacement
import dotenv

dotenv.load_dotenv()


async def main():
    replacer = Replacer(
        replacements=[
            Replacement(key="{{name}}", value="Anmol"),
            Replacement(key="{{age}}", value="23"),
        ]
    )
    adaptor = OpenAIAdaptor(
        module=AsyncOpenAI(api_key=os.getenv("OPENAI_API_KEY")),
        database=ClickHouse(
            host="localhost",
            port=18123,
            user="default",
            password="pass",
            database="cachelm",
            vectorizer=FastEmbedVectorizer(),
        ),
        middlewares=[replacer],
        distance_threshold=0.1,
    )

    openai_adapted = adaptor.get_adapted()

    system_message = {
        "role": "developer",
        "content": "use my name Anmol and age 23 in your responses. ",
    }
    messages = [system_message]

    while True:
        user_input = input("\nAsk your question (or type 'exit' to quit): ")
        if user_input.strip().lower() == "exit":
            break

        messages.append({"role": "user", "content": user_input})

        start_time = time.time()
        completion = await openai_adapted.chat.completions.create(
            model="gpt-4o",
            messages=messages,
        )
        end_time = time.time()

        assistant_message = completion.choices[0].message.content
        print("\nResponse:")
        print(assistant_message)
        print(f"Time taken: {end_time - start_time:.2f} seconds")

        messages.append({"role": "assistant", "content": assistant_message})


asyncio.run(main())



import json


class ToolCall:
    """
    Class to represent a tool call in the chat history.
    """

    def __init__(self, tool: str, args: list):
        self.tool = tool
        self.args = args

    def __repr__(self):
        return f"ToolCall(tool={self.tool}, args={self.args})"

    def to_json_str(self):
        """
        Convert the tool call to a JSON string.
        """
        return json.dumps({"tool": self.tool, "args": self.args})

    def to_json(self):
        """
        Convert the tool call to a JSON object.
        """
        return {"tool": self.tool, "args": self.args}

    @staticmethod
    def from_json_str(json_str: str):
        """
        Create a ToolCall object from a JSON string.
        """
        data = json.loads(json_str)
        return ToolCall(tool=data.get("tool", ""), args=data.get("args", []))

    @staticmethod
    def from_json(json_obj: dict):
        """
        Create a ToolCall object from a JSON object.
        """
        return ToolCall(tool=json_obj.get("tool", ""), args=json_obj.get("args", []))


class Message:
    """
    Class to represent a message in the chat history.
    """

    def __init__(
        self, role: str, content: str, tool_calls: list[ToolCall] | None = None
    ):
        self.role = role
        self.content = content
        self.tool_calls = tool_calls

    def __repr__(self):
        return f"Message(role={self.role}, content={self.content})"

    def to_json_str(self):
        """
        Convert the message to a JSON string.
        """
        return json.dumps(
            {
                "role": self.role,
                "content": self.content,
                "tool_calls": [
                    tool_call.to_json() for tool_call in (self.tool_calls or [])
                ],
            },
        )

    def to_formatted_str(self):
        """
        Convert the message to a formatted string.
        example:
        >>> message = Message(role="user", content="Hello")
        >>> message.to_formatted_str()
        'user: Hello'
        >>> message = Message(role="assistant", content="Hi there!", tool_calls=[{"tool": "calculator", "args": [1, 2]}])
        >>> message.to_formatted_str()
        'assistant: Hi there! (Tool calls: [{"tool": "calculator", "args": [1, 2]}])'
        """
        if self.content != "" and self.tool_calls is None:
            return f"{self.role}: {self.content}"
        elif self.content == "":
            return f""

        return f"{self.role}: {self.content} (Tool calls: {json.dumps([tool_call.to_json() for tool_call in (self.tool_calls or [])])})"

    @staticmethod
    def from_json_str(json_str: str):
        """
        Create a Message object from a JSON string.
        """
        data = json.loads(json_str)
        return Message(
            role=data.get("role", ""),
            content=data.get("content", ""),
            tool_calls=[
                ToolCall.from_json(tool_call)
                for tool_call in data.get("tool_calls", [])
            ],
        )


class ChatHistory:
    """
    Class to represent the chat history.
    """

    def __init__(self):
        self.messages: list[Message] = []

    def add_user_message(self, message: str):
        """
        Add a user message to the chat history.
        """
        self.messages.append(Message(role="user", content=message))

    def add_assistant_message(self, message: str):
        """
        Add an assistant message to the chat history.
        """
        self.messages.append(Message(role="assistant", content=message))

    def set_messages(self, messages: list[Message]):
        """
        Set the messages in the chat history.
        """
        self.messages = messages

    def get_messages(self, length: int = 0) -> list[Message]:
        """
        Get the messages from the chat history.
        If length is 0, return all messages.
        If length is greater than the number of messages, return the last 'length' messages.
        If length is less than the number of messages, return the first 'length' messages. Prepend with empty strings if necessary.
        Example:
        >>> chat_history = ChatHistory()
        >>> chat_history.add_user_message("Hello")
        >>> chat_history.add_assistant_message("Hi there!")
        >>> chat_history.get_messages()
        [Message(role='assistant', content='Hi there!'), Message(role='user', content='Hello')]
        >>> chat_history.get_messages(1)
        [Message(role='assistant', content='Hi there!')]
        >>> chat_history.get_messages(3)
        [Message(role='', content=''), Message(role='assistant', content='Hi there!'), Message(role='user', content='Hello')]
        >>> chat_history.get_messages(4)
        [Message(role='', content=''), Message(role='', content=''), Message(role='assistant', content='Hi there!'), Message(role='user', content='Hello')]
        """
        if length == 0:
            length = len(self.messages)
        if length > len(self.messages):
            # Pad with empty strings at the beginning
            pad = [Message("", "")] * (length - len(self.messages))
            return pad + self.messages
        else:
            return self.messages[-length:]

    def remove_message(self, index: int):
        """
        Remove a message from the chat history.
        """
        if 0 <= index < len(self.messages):
            del self.messages[index]
        else:
            raise IndexError("Index out of range")

    def to_formatted_str(self):
        """
        Convert the chat history to a formatted string.
        """
        return "\n".join([message.to_formatted_str() for message in self.messages])

    def clear(self):
        """
        Clear the chat history.
        """
        self.messages = []

    def __len__(self):
        """
        Get the length of the chat history.
        """
        return len(self.messages)

    def __getitem__(self, index: int):
        """
        Get an item from the chat history.
        """
        return self.messages[index]
